class MatrixError(Exception):
    def __init__(self,msg):
        self.message = msg
    def __str__(self):
        return self.message

class DimensionError(MatrixError):
    def __init__(self,err,*args):
        self.message  = err

class NotListOrTuple(MatrixError):
    """
    A list or a tuple is required
    """
    def __init__(self,err):
        self.message = f"Given value should be a list or a tuple, not '{type(err)}'"

class EmptyMatrix(MatrixError):
    """
    Matrix is empty
    """
    def __init__(self,err,*args):
        self.message  = err

class InvalidColumn(MatrixError):
    """
    Invalid column name
    """
    def __init__(self,err,*args):
        self.message  = f"'{type(err)}' type index '{err}' can't be used as a column index. "+". ".join(args)

class InvalidIndex(MatrixError):
    """
    Invalid row index
    """
    def __init__(self,err,*args):
        self.message  = f"'{type(err)}' type index '{err}' can't be used as a row index. "+". ".join(args)

class FillError(MatrixError):
    """
    Error filling matrices
    """
    def __init__(self,err,*args):
        self.message  = f"'{type(err)}' type '{err}'can't be used to fill matrices. "+". ".join(args)

class DtypeError(MatrixError):
    """
    dtype error
    """
    def __init__(self,err,*args):
        self.message = f"'{err}' isn't a valid value for dtype.\ndtype only accepts following values : 'integer'|'float'|'complex'|'dataframe'. "+". ".join(args)

class ColdtypeError(MatrixError):
    """
    Invalid column dtype
    """
    def __init__(self,err,*args):
        self.message = f"'{type(err)}' type '{err}' can't be used as column dtype. \ncoldtypes should be all 'type' objects. "+". ".join(args)

class InconsistentValues(MatrixError):
    """
    Different value types in a column
    """
    def __init__(self,err,*args):
        self.message  = err

class InvalidList(MatrixError):
    """
    Invalid values to set for coldtypes,features etc.
    """
    def __init__(self,err,req,typ,*args):
        self.message = f"{type(err)}' type '{err}' can't be used as a length {req} list to replace {typ}. "+". ".join(args)
